import pytest
from collections import namedtuple
from configparser import ConfigParser
from unittest.mock import patch, ANY
from pathlib import Path
from click import Context
from click.exceptions import UsageError
from click.testing import CliRunner
from braulio.git import Tag
from braulio.version import Version
from braulio.cli import (
    cli,
    release,
    current_version_option_validator,
    label_pattern_option_validator,
    changelog_file_option_validator,
    bump_option_validator,
    tag_pattern_option_validator,
    message_option_validator,
)

parametrize = pytest.mark.parametrize

FakeTag = namedtuple("FakeTag", ["name"])


@pytest.fixture
def ctx():
    return Context(release)


@parametrize(
    "tag_list, from_arg",
    [
        ([], None),
        ([FakeTag(name="v0.2.0")], "v0.2.0"),
        ([FakeTag(name="v0.3.3"), FakeTag(name="v0.3.2")], "v0.3.3"),
    ],
)
@patch("braulio.cli.Git", autospec=True)
def test_call_to_git_log_method(MockGit, tag_list, from_arg):
    mock_git = MockGit()
    mock_git.tags = tag_list
    mock_git.log.return_value = []
    runner = CliRunner()

    result = runner.invoke(cli, ["release"])

    assert result.exit_code == 0
    mock_git.log.assert_called_with(_from=from_arg)


@patch("braulio.cli.Git", autospec=True)
def test_commitless_repository(MockGit):
    mock_git = MockGit()
    mock_git.log.return_value = []
    runner = CliRunner()

    result = runner.invoke(cli, ["release"])

    assert " › Nothing to release." in result.output

    mock_git.commit.assert_not_called()
    mock_git.tag.assert_not_called()


@parametrize("_input", ["y", "n"])
@patch("braulio.cli.update_chglog")
@patch("braulio.cli.Git", autospec=True)
def test_confirmation_prompt(MockGit, mock_update_chglog, _input, isolated_filesystem):

    runner = CliRunner()

    with isolated_filesystem("HISTORY.rst"):
        result = runner.invoke(cli, ["release"], input=_input)

        assert result.exit_code == 0
        assert " › Continue? [y/N]" in result.output

        mock_git = MockGit()

        called = True if _input == "y" else False
        assert mock_update_chglog.called is called
        assert mock_git.commit.called is called
        assert mock_git.tag.called is called


@parametrize(
    "value, expected",
    [("6", Version("6.0.0")), ("4.3", Version("4.3.0")), ("3.2.1", Version("3.2.1"))],
)
def test_bump_option_validator(ctx, value, expected):
    assert bump_option_validator(ctx, {}, value) == expected


def test_bump_option_validator_with_invalid_string(ctx):
    with pytest.raises(UsageError):
        bump_option_validator(ctx, {}, "invalid-version-string")


@parametrize(
    "option, tags, current_version, expected",
    [
        ("--major", [FakeTag("v0.10.1")], Version("0.10.1"), "1.0.0"),
        ("--minor", [FakeTag("v0.10.1")], Version("0.10.1"), "0.11.0"),
        ("--patch", [FakeTag("v0.10.1")], Version("0.10.1"), "0.10.2"),
        ("--major", [], Version(), "1.0.0"),
        ("--minor", [], Version(), "0.1.0"),
        ("--patch", [], Version(), "0.0.1"),
        ("--bump=4.0.0", [FakeTag("v2.0.1")], Version("2.0.1"), "4.0.0"),
        ("--bump=3.0.0", [], Version(), "3.0.0"),
    ],
)
@patch("braulio.cli.update_chglog", autospec=True)
@patch("braulio.cli.Git", autospec=True)
def test_manual_version_bump(
    MockGit,
    mock_update_chglog,
    option,
    tags,
    current_version,
    expected,
    isolated_filesystem,
):

    mock_git = MockGit()
    mock_git.tags = tags
    runner = CliRunner()

    with isolated_filesystem("HISTORY.rst"):
        result = runner.invoke(cli, ["release", option], input="y")

        assert result.exit_code == 0

        # Check what version was passed to update_chglog function
        mock_update_chglog.assert_called_with(
            Path("HISTORY.rst"),
            current_version=current_version,
            new_version=Version(expected),
            release_data={},
        )

        mock_git.commit.assert_called_with(
            f"Release version {expected}", files=["HISTORY.rst"]
        )
        mock_git.tag.assert_called_with(f"v{expected}")


@patch("braulio.cli.Git", autospec=True)
def test_bump_to_a_lower_version(MockGit, isolated_filesystem):

    mock_git = MockGit()
    mock_git.tags = [FakeTag("v2.0.0")]
    runner = CliRunner()

    with isolated_filesystem("HISTORY.rst"):
        result = runner.invoke(cli, ["release", "--bump=1", "-y"])

    assert result.exit_code == 1
    assert "The release of a lower versions is not supported for now" in result.output
    assert "Aborted" in result.output


@parametrize(
    "hash_lst, tags, expected",
    [
        (["4d17c1a", "80a9e0e"], [FakeTag("v0.7.6")], "0.7.7"),
        (["4d17c1a", "ccaa185"], [FakeTag("v0.7.6")], "0.8.0"),
        (["8c8dcb7", "ccaa185"], [FakeTag("v0.7.6")], "1.0.0"),
        (["4d17c1a", "80a9e0e"], [], "0.0.1"),
        (["4d17c1a", "ccaa185"], [], "0.1.0"),
        (["8c8dcb7", "ccaa185"], [], "1.0.0"),
    ],
)
@patch("braulio.cli.update_chglog", autospec=True)
@patch("braulio.cli.Git", autospec=True)
def test_determine_next_version_from_commit_messages(
    MockGit,
    mocked_update_changelog,
    hash_lst,
    tags,
    expected,
    commit_registry,
    isolated_filesystem,
):
    mock_git = MockGit()
    mock_git.tags = tags
    mock_git.log.return_value = [commit_registry[short_hash] for short_hash in hash_lst]

    runner = CliRunner()

    with isolated_filesystem("HISTORY.rst"):
        result = runner.invoke(cli, ["release", "-y"])

        assert result.exit_code == 0, result.exception

        mocked_update_changelog.assert_called()

        # Check what version was passed to update_chglog function
        passed_args = mocked_update_changelog.call_args[1]
        assert passed_args["new_version"] == Version(expected)

        mock_git.commit.assert_called_with(
            f"Release version {expected}", files=["HISTORY.rst"]
        )
        mock_git.tag.assert_called_with(f"v{expected}")


@patch("braulio.cli.Git", autospec=True)
def test_release_from_pre_release_stage(MockGit, commit_list, isolated_filesystem):
    runner = CliRunner()
    mock_git = MockGit()
    mock_git.tags = [FakeTag("v1.3.1beta5")]
    mock_git.log.return_value = commit_list

    with isolated_filesystem("HISTORY.rst"):
        # Add first pre-release stages to config file
        cfg = Path("setup.cfg")
        cfg.touch()
        cfg.write_text(
            "[braulio.stages]\n"
            "dev   = {major}.{minor}.{patch}.dev{n}\n"
            "beta  = {major}.{minor}.{patch}beta{n}\n"
            "final = {major}.{minor}.{patch}\n"
        )

        result = runner.invoke(cli, ["release"])

        assert result.exit_code == 0, result.output
        assert "New version      : 1.3.1" in result.output


@patch("braulio.cli.ReleaseDataTree")
@patch("braulio.cli.commit_analyzer")
@patch("braulio.cli.update_chglog", autospec=True)
@patch("braulio.cli.get_next_version")
@patch("braulio.cli.Git", autospec=True)
def test_call_to_update_changelog(
    MockGit,
    mock_get_next_version,
    mock_update_chglog,
    mock_commit_analyzer,
    MockReleaseDataTree,
    isolated_filesystem,
):
    runner = CliRunner()
    mock_git = MockGit()
    mock_git.tags = []

    with isolated_filesystem("HISTORY.rst"):
        result = runner.invoke(cli, ["release", "-y"])

        assert result.exit_code == 0

        mock_git.log.assert_called()

        mock_commit_analyzer.assert_called_with(
            mock_git.log(), "!{action}:{scope}", "footer"
        )

        MockReleaseDataTree.assert_called_with(mock_commit_analyzer())

        release_data = MockReleaseDataTree()

        mock_get_next_version.assert_called_with(
            Version(), release_data.bump_version_to
        )

        mock_update_chglog.assert_called_with(
            Path("HISTORY.rst"),
            current_version=Version(),
            new_version=mock_get_next_version(),
            release_data=release_data,
        )


def test_changelog_file_option_validator(ctx, isolated_filesystem):
    with isolated_filesystem("FILE"):
        path = changelog_file_option_validator(ctx, {}, "FILE")

        assert path == Path("FILE")


def test_changelog_file_option_validator_with_missing_file(ctx, isolated_filesystem):

    with isolated_filesystem:
        with pytest.raises(UsageError):
            changelog_file_option_validator(ctx, {}, "MISSING_FILE")


@patch("braulio.cli.Git", autospec=True)
def test_changelog_not_found(MockGit):
    runner = CliRunner()

    with runner.isolated_filesystem():
        result = runner.invoke(cli, ["release", "-y"])

        assert result.exit_code == 2
        assert "Unable to find HISTORY.rst" in result.output
        assert 'Run "$ brau init" to create one' in result.output

        mock_git = MockGit()
        mock_git.commit.assert_not_called()
        mock_git.tag.assert_not_called()


@patch("braulio.cli.Git", autospec=True)
@patch("braulio.cli.update_files", autospec=True)
def test_files_argument_from_command_line(mock_update_files, MockGit, fake_repository):

    runner = CliRunner()
    mock_git = MockGit()
    mock_git.tags = []

    with fake_repository("black"):
        files = ["black/__init__.py", "setup.py"]
        result = runner.invoke(cli, ["release", "-y"] + files)

    assert result.exit_code == 0, result.exception
    mock_update_files.assert_called_with(
        ("black/__init__.py", "setup.py"), "0.0.0", "0.0.1"
    )


@patch("braulio.cli.Git", autospec=True)
@patch("braulio.cli.update_files", autospec=True)
def test_files_argument_from_config_file(mock_update_files, MockGit, fake_repository):

    runner = CliRunner()
    mock_git = MockGit()
    mock_git.tags = []

    with fake_repository("white"):
        result = runner.invoke(cli, ["release", "-y"])

    assert result.exit_code == 0
    mock_update_files.assert_called_with(
        ("white/__init__.py", "setup.py"), "0.0.0", "0.0.1"
    )


@patch("braulio.cli.Git", autospec=True)
@patch("braulio.cli.update_files", autospec=True)
def test_added_files_to_release_commit(mock_update_files, MockGit, fake_repository):

    runner = CliRunner()
    mock_git = MockGit()
    mock_git.tags = []

    with fake_repository("white"):
        result = runner.invoke(cli, ["release", "--commit", "-y"])

    assert result.exit_code == 0
    mock_git.commit.assert_called_with(
        "Release version 0.0.1", files=["HISTORY.rst", "white/__init__.py", "setup.py"]
    )


@parametrize("options", [["--no-commit", "--bump=0.2.1"], ["--no-tag"]])
@patch("braulio.cli.Git", autospec=True)
def test_output_after_confirmation_prompt(
    MockGit, isolated_filesystem, commit_list, options
):

    runner = CliRunner()
    mock_git = MockGit()
    mock_git.tags = [FakeTag("v0.2.0")]
    mock_git.log.return_value = commit_list

    with isolated_filesystem("HISTORY.rst"):
        result = runner.invoke(cli, ["release"] + options)

        assert result.exit_code == 0, result.output
        assert " › Current version  : 0.2.0" in result.output
        assert f" › Commits found    : {len(commit_list)}" in result.output
        assert " › New version      :" in result.output
        assert " › Braulio will perform the next tasks :" in result.output

        if "--no-tag" not in options:
            assert "Tag the repository with v0.2.1" in result.output

        if "--no-commit" not in options:
            assert "Add a release commit" in result.output

        assert " › Continue?" in result.output


@parametrize("options", [["--no-commit", "--bump=0.2.1"], ["--no-tag", "--bump=0.2.1"]])
@patch("braulio.cli.Git", autospec=True)
def test_output_before_confirmation_prompt(
    MockGit, isolated_filesystem, commit_list, options
):

    runner = CliRunner()
    mock_git = MockGit()
    mock_git.tags = [Tag("v0.2.0")]
    mock_git.log.return_value = commit_list

    with isolated_filesystem("HISTORY.rst"):
        result = runner.invoke(cli, ["release"] + options, input="y")

        assert result.exit_code == 0, result.output
        assert "Update changelog ✓" in result.output

        if "--no-tag" not in options:
            assert " › Add tag v0.2.1 ✓" in result.output

        if "--no-commit" not in options:
            assert " › Add commit: Release version 0.2.1 ✓" in result.output

        assert "Version 0.2.1 released successfully 🎉" in result.output


@parametrize(
    "label_position, value",
    [
        ("header", ""),
        ("header", "{subject}"),
        ("header", "{action}:{scope}"),
        ("footer", "{subject}"),
    ],
)
def test_invalid_label_pattern_option(ctx, label_position, value):
    ctx.params["label_position"] = label_position

    with pytest.raises(UsageError):
        label_pattern_option_validator(ctx, {}, value)


@parametrize(
    "cfg, option",
    [
        ({"label_pattern": "{action}:{scope}"}, []),
        ({}, ["--label-pattern={action}:{scope}"]),
    ],
)
@patch("braulio.cli.commit_analyzer", autospec=True)
@patch("braulio.cli.Git", autospec=True)
def test_label_pattern_option(
    MockGit, mock_commit_analyzer, cfg, option, isolated_filesystem
):

    runner = CliRunner()

    with isolated_filesystem("HISTORY.rst", cfg=cfg):
        result = runner.invoke(cli, ["release"] + option)

        assert result.exit_code == 0
        mock_commit_analyzer.assert_called_with(
            MockGit().log(), "{action}:{scope}", "footer"
        )


@parametrize(
    "flag, cfg, called",
    [
        ([], {}, True),
        ([], {"commit": False}, False),
        (["--no-commit"], {"commit": True}, False),
        (["--commit"], {"commit": False}, True),
    ],
)
@patch("braulio.cli.Git", autospec=True)
def test_commit_flag(MockGit, flag, cfg, called, isolated_filesystem):
    """Test commit flag picked from CLI or configuration file"""

    mock_git = MockGit()
    runner = CliRunner()

    with isolated_filesystem("HISTORY.rst", cfg=cfg):
        command = ["release"] + flag + ["-y"]
        result = runner.invoke(cli, command)

    assert result.exit_code == 0
    assert mock_git.tag.called is True
    assert mock_git.commit.called is called


@parametrize(
    "flag, cfg, called",
    [
        ([], {}, True),
        ([], {"tag": False}, False),
        (["--no-tag"], {"tag": True}, False),
        (["--tag"], {"tag": False}, True),
    ],
)
@patch("braulio.cli.Git", autospec=True)
def test_tag_flag(MockGit, flag, cfg, called, isolated_filesystem):
    """Test tag flag picked from CLI or configuration file"""

    mock_git = MockGit()
    runner = CliRunner()

    with isolated_filesystem("HISTORY.rst", cfg=cfg):
        command = ["release"] + flag + ["-y"]
        result = runner.invoke(cli, command)

    assert result.exit_code == 0
    assert mock_git.tag.called is called
    assert mock_git.commit.called is True


@parametrize(
    "value",
    (
        "Bump to {new_version}",
        "Bump from {current_version} to {new_version}",
        "{new_version}",
    ),
)
def test_message_option_validator(ctx, value):
    rv = message_option_validator(ctx, {}, value)
    assert rv == value


@parametrize("value", (None, "message"))
def test_message_option_validator_with_invalid_value(ctx, value):
    with pytest.raises(UsageError):
        message_option_validator(ctx, {}, "message")


@parametrize(
    "cfg, option, expected",
    [
        ({}, [], "Release version 8.0.1"),
        ({"message": "Bump {new_version}"}, [], "Bump 8.0.1"),
        ({}, ["--message=Bump {new_version}"], "Bump 8.0.1"),
        ({"message": "{current_version} => {new_version}"}, [], "8.0.0 => 8.0.1"),
    ],
)
@patch("braulio.cli.Git")
def test_message_option(MockGit, isolated_filesystem, cfg, option, expected):
    mock_git = MockGit()
    mock_git.tags = [FakeTag(name="v8.0.0")]
    runner = CliRunner()

    with isolated_filesystem("HISTORY.rst", cfg=cfg):
        command = ["release", "-y"] + option
        result = runner.invoke(cli, command)

        assert result.exit_code == 0

        mock_git.commit.assert_called_with(expected, files=ANY)


@parametrize("value", ("v{version}", "release{version}", "{version}"))
def test_tag_pattern_option_validator(ctx, value):
    assert tag_pattern_option_validator(ctx, {}, value) == value


def test_tag_pattern_option_validator_with_invalid_value(ctx):
    with pytest.raises(UsageError):
        tag_pattern_option_validator(ctx, {}, "invalid-tag-pattern")


@parametrize(
    "options, cfg, expected",
    [
        ([], {}, "v0.0.1"),
        ([], {"tag_pattern": "version{version}"}, "version0.0.1"),
        ([], {"tag_pattern": "release-{version}"}, "release-0.0.1"),
        (["--tag-pattern={version}-released"], {}, "0.0.1-released"),
        (
            ["--tag-pattern=version{version}"],
            {"tag_pattern": "release-{version}"},
            "version0.0.1",
        ),
    ],
)
@patch("braulio.cli.Git", autospec=True)
def test_tag_pattern_option(MockGit, isolated_filesystem, options, cfg, expected):
    """Test tag options picked from CLI or configuration file"""

    mock_git = MockGit()
    mock_git.tags = []
    runner = CliRunner()

    with isolated_filesystem("HISTORY.rst", cfg=cfg):
        command = ["release"] + options + ["-y"]
        result = runner.invoke(cli, command)

    assert result.exit_code == 0
    mock_git.tag.assert_called_with(expected)


def test_current_version_option_validator_with_invalid_version_string(ctx):
    with pytest.raises(UsageError):
        current_version_option_validator(ctx, {}, "invalid-version-string")


@parametrize(
    "tags, value, expected_version, expected_tag",
    [
        ([], None, None, None),
        ([], "0.12.0", Version("0.12.0"), None),
        (
            [FakeTag("v2.0.0"), FakeTag("v1.9.10")],
            None,
            Version("2.0.0"),
            FakeTag("v2.0.0"),
        ),
        ([FakeTag("v2.0.0")], "3.0.0", Version("3.0.0"), None),
        ([FakeTag("v4.0.0")], "2.0.0", Version("2.0.0"), None),
        ([FakeTag("save-point")], "3.0.0", Version("3.0.0"), None),
        ([FakeTag("save-point")], None, None, None),
    ],
)
@patch("braulio.cli.Git")
def test_current_version_option_validator(
    MockGit, tags, value, expected_version, expected_tag
):
    mock_git = MockGit()
    mock_git.tags = tags
    ctx = Context(release)
    ctx.params["tag_pattern"] = "v{version}"

    version = current_version_option_validator(ctx, {}, value)

    assert version == expected_version

    if expected_tag:
        assert ctx.params["current_tag"] == expected_tag


@parametrize(
    "cfg, options, expected",
    [
        ({"current_version": "2.0.0"}, [], "2.0.1"),
        ({"current_version": "2.0.0"}, ["--current-version=2.0.0"], "2.0.1"),
        ({}, ["--current-version=2.0.0"], None),
        ({}, [], None),
    ],
)
@patch("braulio.cli.Git")
def test_update_current_version_config_file_option(
    MockGit, cfg, options, expected, isolated_filesystem
):

    mock_git = MockGit()
    mock_git.tags = [FakeTag(name="v2.0.0")]
    runner = CliRunner()

    with isolated_filesystem("HISTORY.rst", cfg=cfg):

        command = ["release", "-y"] + options
        result = runner.invoke(cli, command)

        assert result.exit_code == 0

        cfg = ConfigParser()
        cfg.read("setup.cfg")



import os
import pytest
from configparser import ConfigParser
from shutil import copytree
from pathlib import Path
from braulio.git import _extract_commit_texts, Commit


class IsolatedFilesystem:
    def __init__(self, tmpdir):
        self.tmpdir = tmpdir
        self.chglog = None
        self.cfg = None

    def __enter__(self):
        self.original_dir = self.tmpdir.chdir()

        if self.chglog:
            Path(self.chglog).touch()

        if self.cfg:
            with open("setup.cfg", "w") as config_file:
                config_parser = ConfigParser()
                config_parser["braulio"] = self.cfg
                config_parser.write(config_file)

        return self

    def __exit__(self, *args):
        os.chdir(self.original_dir)

    def __call__(self, chglog=None, cfg=None):
        self.chglog = chglog
        self.cfg = cfg

        return self


class FakeRepository:

    repository_dir = Path.cwd() / "tests" / "repos"

    def __init__(self, tmpdir, repository_name):
        self.tmpdir = tmpdir
        self.temporal_dir = Path(tmpdir) / "repo"
        self.original_dir = Path.cwd()
        self.repository_name = repository_name

    def __enter__(self):
        # Copy fake repository to temporal directory
        copytree(src=self.repository_dir / self.repository_name, dst=self.temporal_dir)

        # Switch to temporal directory
        os.chdir(self.temporal_dir)

    def __exit__(self, *args):
        # Switch back to original directory
        os.chdir(self.original_dir)


@pytest.fixture
def fake_repository(tmpdir):
    def _repository(repository_name):
        return FakeRepository(tmpdir, repository_name)

    return _repository


@pytest.fixture
def isolated_filesystem(tmpdir):
    return IsolatedFilesystem(tmpdir)


@pytest.fixture(scope="session")
def fake_git_log_output():
    return Path("tests/data/commits.txt").read_text()


@pytest.fixture(scope="session")
def commit_text_list(fake_git_log_output):
    return _extract_commit_texts(fake_git_log_output)


@pytest.fixture(scope="session")
def commit_list(commit_text_list):
    return [Commit(text) for text in commit_text_list]


@pytest.fixture(scope="session")
def commit_text_registry(commit_text_list):
    return {c[7:14]: c for c in commit_text_list}


@pytest.fixture(scope="session")
def commit_registry(commit_list):
    return {commit.hash[:7]: commit for commit in commit_list}



import pytest
from configparser import ConfigParser
from collections import OrderedDict
from pathlib import Path
from click.testing import CliRunner
from braulio.config import Config, update_config_file


parametrize = pytest.mark.parametrize


class TestConfig:
    def check_default_options(self, config):
        assert config.tag is True
        assert config.commit is True
        assert config.message == "Release version {new_version}"
        assert config.confirm is False
        assert config.files == ()
        assert config.changelog_file == Path("HISTORY.rst")
        assert config.label_position == "footer"
        assert config.label_pattern == "!{action}:{scope}"
        assert config.tag_pattern == "v{version}"
        assert config.current_version is None
        assert config.stages == {"final": "{major}.{minor}.{patch}"}

    def test_default_options(self, isolated_filesystem):

        with isolated_filesystem:
            config = Config()
            self.check_default_options(config)

    def test_cfg_file_property(self, isolated_filesystem):

        with isolated_filesystem:
            file_path = Path.cwd() / "setup.cfg"
            file_path.write_text(
                "[braulio]\n" "option1 = value1\n" "option2 = value2\n"
            )

            config = Config()
            config.cfg_file_options == {"option1": "value1", "option2": "value2"}

    def test_config_file_with_not_braulio_section(self, isolated_filesystem):
        config_file_content = "[section]\n" "option1 = value1\n" "option2 = value2\n"

        with isolated_filesystem:
            file_path = Path.cwd() / "setup.cfg"
            file_path.write_text(config_file_content)

            config = Config()
            self.check_default_options(config)

    def test_merge_options_from_config_file(self, isolated_filesystem):
        config_file_content = (
            "[braulio]\n" "commit = False\n" "tag = False\n" "confirm = True\n"
        )

        with isolated_filesystem:
            file_path = Path.cwd() / "setup.cfg"
            file_path.write_text(config_file_content)

            config = Config()

        assert config.tag is False
        assert config.commit is False
        assert config.confirm is True
        assert config.files == ()

    @parametrize(
        "files_value, expected",
        [
            ("", ()),
            (" ", ()),
            ("folder/file.py", ("folder/file.py",)),
            (" folder/file.py, file.py ", ("folder/file.py", "file.py")),
            (
                " folder/file.py , \n    folder/module.py ,\n    file.py  ",
                ("folder/file.py", "folder/module.py", "file.py"),
            ),
        ],
        ids=[
            "empty value",
            "empty space",
            "single file",
            "multiple files in a single line",
            "multiple files in multiple lines",
        ],
    )
    def test_files_option(self, isolated_filesystem, files_value, expected):
        config_file_content = "[braulio]\n" f"files = {files_value}\n"
        with isolated_filesystem:
            file_path = Path.cwd() / "setup.cfg"
            file_path.write_text(config_file_content)

            config = Config()

        assert type(config.files) == tuple
        assert config.files == expected

    @parametrize(
        "options, expected",
        [
            ({}, {"tag": True, "commit": True, "confirm": False}),
            ({"tag": False}, {"tag": False, "commit": True, "confirm": False}),
            ({"commit": False}, {"tag": True, "commit": False, "confirm": False}),
            ({"confirm": True}, {"tag": True, "commit": True, "confirm": True}),
        ],
    )
    def test_merge_config_file_options_on_default_options(
        self, isolated_filesystem, options, expected
    ):
        """Test to make sure that the options in configuration files have
        precedence over the default configuration.
        """

        with isolated_filesystem:

            # Write a config file to the file system
            with open("setup.cfg", "w") as config_file:
                config_parser = ConfigParser()
                config_parser["braulio"] = options
                config_parser.write(config_file)

            config = Config()

        assert config.tag is expected["tag"]
        assert config.commit is expected["commit"]
        assert config.confirm is expected["confirm"]

    def test_current_version_option(self, isolated_filesystem):
        with isolated_filesystem:
            file_path = Path.cwd() / "setup.cfg"
            file_path.write_text("[braulio]\n" "current_version = 4.9.3\n")

            config = Config()
            assert config.current_version == "4.9.3"

    def test_message_option(self, isolated_filesystem):
        with isolated_filesystem:
            file_path = Path.cwd() / "setup.cfg"
            file_path.write_text("[braulio]\n" "message = Release: {new_version}\n")

            config = Config()
            assert config.message == "Release: {new_version}"

    @pytest.mark.wip
    def test_stages_option(self, isolated_filesystem):
        with isolated_filesystem:
            file_path = Path.cwd() / "setup.cfg"
            file_path.write_text(
                "[braulio.stages]\n"
                "dev   = {major}.{minor}.{patch}.dev{n}\n"
                "beta  = {major}.{minor}.{patch}b{n}\n"
                "final = {major}.{minor}.{patch}\n"
            )

            config = Config()
            assert type(config.stages) == OrderedDict
            assert config.stages == OrderedDict(
                {
                    "dev": "{major}.{minor}.{patch}.dev{n}",
                    "beta": "{major}.{minor}.{patch}b{n}",
                    "final": "{major}.{minor}.{patch}",
                }
            )


class TestUpdateConfigFile:
    def test_empty_directory(self, isolated_filesystem):
        runner = CliRunner()

        with runner.isolated_filesystem():
            setup_cfg_path = Path.cwd() / "setup.cfg"

            update_config_file("option1", "value1")

            assert setup_cfg_path.exists()

            config_parser = ConfigParser()
            config_parser.read("setup.cfg")
            assert config_parser.has_section("braulio")
            assert config_parser.get("braulio", "option1") == "value1"

    def test_setup_cfg_exists(self, isolated_filesystem):
        runner = CliRunner()

        with runner.isolated_filesystem():
            setup_cfg_path = Path.cwd() / "setup.cfg"
            setup_cfg_path.write_text(
                "[section1]\n" "option1 = value1\n" "option2 = value2\n"
            )

            update_config_file("option3", "value3")

            config_parser = ConfigParser()
            config_parser.read("setup.cfg")
            assert config_parser.has_section("braulio")
            assert config_parser.get("braulio", "option3") == "value3"



import re
from collections import OrderedDict
from string import Formatter


VERSION_STRING_PATTERN = (
    "(?P<major>\d+)"  # Major part
    "(?:\.(?P<minor>\d+))?"  # Minor part (optional)
    "(?:\.(?P<patch>\d+))?"  # Patch part (optional)
    "(?:\.?(?P<stage>[a-z]+)(?P<n>\d+))?"  # Pre-release (optional)
)

VERSION_STRING_REGEXP = re.compile(f"^{VERSION_STRING_PATTERN}$")


def validate_version_str(string):
    return bool(re.match(VERSION_STRING_REGEXP, string))


def parse_version_string_parts(string):
    match = VERSION_STRING_REGEXP.match(string)

    if not match:
        return None

    parts = match.groupdict(default=0)

    return OrderedDict(
        major=int(parts["major"]),
        minor=int(parts["minor"]),
        patch=int(parts["patch"]),
        stage=parts["stage"] or None,
        n=int(parts["n"]),
    )


stage_pattern = re.compile(
    "^(?:{major}\.{minor}\.{patch})"  # Required parts
    "(?:\.?(?P<stage>[a-z]+)?\{n\})?$"  # Optional parts
)


class Version:

    _stages = {"final": "{major}.{minor}.{patch}"}

    def __init__(self, string=None, major=0, minor=0, patch=0, stage=None, n=0):

        if string:
            parts = parse_version_string_parts(string)

            if not parts:
                raise ValueError(f"Invalid version string: {string}")

            major, minor, patch, stage, n = parts.values()

        stage = stage or "final"

        try:
            self.serializer = self._stages[stage]
            index = tuple(self._stages).index(stage)
        except KeyError:
            raise ValueError(f"{stage} is an unknown stage")

        self.major = int(major)
        self.minor = int(minor)
        self.patch = int(patch)
        self.stage = stage or None
        self.n = int(n)

        args = major, minor, patch, index, n
        self._number = int("{:0>3}{:0>3}{:0>3}{:0>3}{:0>3}".format(*args))

        parts = {}
        serializer = self.serializer

        for p in Formatter().parse(serializer):
            part = p[1]
            if part:
                parts[p[1]] = getattr(self, p[1])

        self._string = serializer.format(**parts)
        self._stage_number = index

    @classmethod
    def set_stages(cls, serializers):
        _stages = OrderedDict()

        for serializer in serializers:
            match = stage_pattern.match(serializer)

            if not match:
                raise ValueError(f"{serializer} is an invalid serializer")

            key = match.groupdict(default="final")["stage"]
            _stages[key] = serializer

        cls._stages = _stages

    def bump(self, bump_part):
        """Return a new bumped version instance."""

        major, minor, patch, stage, n = tuple(self)

        # stage bump
        if bump_part not in {"major", "minor", "patch"}:

            if bump_part not in self._stages:
                raise ValueError(f"Unknown {bump_part} stage")

            # We can not bump from final stage to final again.
            if self.stage == "final" and bump_part == "final":
                raise ValueError(f"{self} is already in final stage.")

            # bump in the same stage (numeric part)
            if bump_part == self.stage:
                n += 1
            else:
                new_stage_number = tuple(self._stages).index(bump_part)

                # We can not bump to a previous stage
                if new_stage_number < self._stage_number:
                    raise ValueError(f"{bump_part} stage is previous to {self}")

                stage = bump_part
                n = 0
        else:
            # major, minor, or patch bump

            # Only version in final stage can do a major, minor or patch
            # bump
            if self.stage != "final":
                raise ValueError(
                    f"{self} is a pre-release version."
                    f" Can't do a {bump_part} version bump"
                )

            if bump_part == "major":
                major += 1
                minor, patch = 0, 0
            elif bump_part == "minor":
                minor += 1
                patch = 0
            else:
                patch += 1

        return Version(major=major, minor=minor, patch=patch, stage=stage, n=n)

    @property
    def string(self):
        return self._string

    def _is_comparable(self, value):
        if not isinstance(value, Version):
            raise TypeError(
                f"Comparison not supported"
                f" between instances of '{type(self)}'"
                f" and '{type(value)}'"
            )

    def __eq__(self, v):
        self._is_comparable(v)

        return self._number == v._number

    def __lt__(self, v):
        self._is_comparable(v)
        return self._number < v._number

    def __le__(self, v):
        self._is_comparable(v)

        return self._number <= v._number

    def __iter__(self):
        yield self.major
        yield self.minor
        yield self.patch
        yield self.stage
        yield self.n

    def __str__(self):
        return self.string

    def __repr__(self):
        return f"Version('{self.string}')"


def get_next_version(current_version, bump_version_to):

    if bump_version_to not in {"major", "minor", "patch"}:
        new_version = Version(bump_version_to)

        if current_version and new_version <= current_version:
            return None

        return new_version

    new_version = Version()

    if current_version:
        new_version = Version(current_version.string)

    return new_version.bump(bump_version_to)



import pytest
from unittest.mock import patch
from collections import namedtuple
from datetime import date
from pathlib import Path
from braulio.version import Version
from braulio.git import commit_analyzer
from braulio.files import (
    DEFAULT_CHANGELOG,
    KNOWN_CHANGELOG_FILES,
    _render_title,
    _render_subtitle,
    _render_list,
    _render_release,
    update_chglog,
    update_files,
    is_title,
    _split_chglog,
    ReleaseDataTree,
)


parametrize = pytest.mark.parametrize


def test_default_changelog():
    assert DEFAULT_CHANGELOG == "HISTORY.rst"


def test_known_changelog_files():
    assert len(KNOWN_CHANGELOG_FILES) == 3
    assert "HISTORY.rst" in KNOWN_CHANGELOG_FILES
    assert "CHANGELOG.rst" in KNOWN_CHANGELOG_FILES
    assert "CHANGES.rst" in KNOWN_CHANGELOG_FILES


class TestReleaseDataTree:

    FakeSemanticCommit = namedtuple(
        "FakeSemanticCommit", ["action", "scope", "message"]
    )

    @pytest.mark.parametrize(
        "semantic_commits, expected",
        [
            (
                [
                    FakeSemanticCommit(action="fix", scope=None, message=""),
                    FakeSemanticCommit(
                        action="refactor", scope=None, message="BREAKING CHANGES"
                    ),
                ],
                "major",
            ),
            (
                [
                    FakeSemanticCommit(action="fix", scope=None, message=""),
                    FakeSemanticCommit(action="feat", scope=None, message=""),
                ],
                "minor",
            ),
            (
                [
                    FakeSemanticCommit(action="fix", scope=None, message=""),
                    FakeSemanticCommit(action="refactor", scope=None, message=""),
                ],
                "patch",
            ),
        ],
    )
    def test_bump_version_to(self, semantic_commits, expected):

        release = ReleaseDataTree(semantic_commits)

        assert release.bump_version_to == expected

    def test_commit_grouping_by_action_and_scope(self, commit_list):
        semantic_commits = commit_analyzer(
            commit_list, label_pattern="!{action}:{scope}"
        )
        release_data = ReleaseDataTree(semantic_commits)

        assert len(release_data["fix"]["thing"]) == 1
        assert len(release_data["feat"]["scopeless"]) == 1
        assert len(release_data["feat"]["cli"]) == 1
        assert len(release_data["feat"]["music"]) == 2
        assert len(release_data["refactor"]["music"]) == 1
        assert len(release_data["refactor"]["lorem"]) == 1


@parametrize(
    "level, expected",
    [(1, "Title\n=====\n\n"), (2, "Title\n-----\n\n"), (3, "Title\n~~~~~\n\n")],
)
def test_render_title(level, expected):
    output = _render_title("Title", level)
    assert output == expected


def test_render_sublist(commit_registry):

    FakeSemanticCommit = namedtuple("FakeSemanticCommit", ["subject"])

    semantic_commits = [
        FakeSemanticCommit(subject="subject 1"),
        FakeSemanticCommit(subject="subject 2"),
        FakeSemanticCommit(subject="subject 3"),
    ]

    markup = _render_subtitle(semantic_commits)

    assert markup == (
        f"  - {semantic_commits[0].subject}\n"
        f"  - {semantic_commits[1].subject}\n"
        f"  - {semantic_commits[2].subject}\n"
    )


def test_render_list(commit_text_registry):

    FakeSemanticCommit = namedtuple("FakeSemanticCommit", ["subject"])

    commits = [
        FakeSemanticCommit(subject="subject 1"),
        FakeSemanticCommit(subject="subject 2"),
        FakeSemanticCommit(subject="subject 3"),
    ]

    release_data_tree = {
        "scopeless": [commits[0], commits[1]],
        "scope1": [commits[0], commits[1]],
        "scope2": [commits[1]],
        "scope3": [commits[0], commits[1]],
    }

    markup = _render_list(release_data_tree)

    assert markup == (
        f"* {commits[0].subject}\n"
        f"* {commits[1].subject}\n"
        "* scope1\n\n"
        f"  - {commits[0].subject}\n"
        f"  - {commits[1].subject}\n"
        f"* scope2 - {commits[1].subject}\n"
        "* scope3\n\n"
        f"  - {commits[0].subject}\n"
        f"  - {commits[1].subject}\n\n"
    )


@parametrize(
    "first, second, third, expected",
    [
        (None, None, None, False),
        (None, None, "-----\n", False),
        (None, "title\n", "-----\n", False),
        ("title\n", "-----\n", "\n", True),
        ("title\n", "-----\n", None, True),
        ("\n", "title\n", "-----\n", True),
        ("-----\n", "title\n", "-----\n", True),
        ("---------\n", "  title  \n", "---------\n", True),
        ("title\n", "-----\n", "Paragraphs.", True),
        ("Paragrah.", "title\n", "-----\n", False),
        ("-----\n", "title\n", "Paragraphs.", False),
    ],
)
def test_is_title(first, second, third, expected):
    assert is_title(first, second, third) is expected


rst_file1 = """
.. currentmodule:: mymodule

Title
=====

-------------
Section title
-------------

Some Other Content

"""

rst_file2 = """
Title
=====

This is a paragraph.

Another paragraph.

Section title
-------------

Some Other Content

"""

rst_file3 = """Section title
~~~~~~~~~~~~~

"""

rst_file4 = "Paragrah 1.\nParagrah 2.\nParagrah 3."

rst_file5 = ""


@parametrize(
    "file_content, endswith, startswith",
    [
        (rst_file1, "=====\n\n", "-------------"),
        (rst_file2, "Another paragraph.\n\n", "Section title"),
        (rst_file3, "", "Section title"),
        (rst_file4, "Paragrah 3.", ""),
        (rst_file5, "", ""),
    ],
    ids=[
        "Main title without paragraph",
        "Main title with paragraph",
        "Empty content",
        "Without title",
        "empty file",
    ],
)
def test_split_chglog(isolated_filesystem, file_content, endswith, startswith):

    with isolated_filesystem:
        path = Path("document.rst")
        path.write_text(file_content)

        top, bottom = _split_chglog(path, "Section title")

        assert top.endswith(endswith)
        assert bottom.startswith(startswith)


class Test_render_release:
    def test_release_with_fixes_and_features(self, commit_list):
        version = Version(major=10, minor=3, patch=0)
        semantic_commits = commit_analyzer(
            commit_list, label_pattern="!{action}:{scope}"
        )
        release_data = ReleaseDataTree(semantic_commits)
        markup = _render_release(version, release_data=release_data)

        assert markup == (
            f"10.3.0 ({str(date.today())})\n"
            "-------------------\n\n"
            "Bug Fixes\n"
            "~~~~~~~~~\n\n"
            "* thing - Fix a thing\n\n"
            "Features\n"
            "~~~~~~~~\n\n"
            "* Add a thing\n"
            "* music\n\n"
            "  - Add more music please\n"
            "  - Add cool musics :D\n"
            "* cli - Add a cli tool\n\n"
        )

    def test_release_with_fixes(self, commit_registry):
        semantic_commits = commit_analyzer(
            [commit_registry["4d17c1a"]], label_pattern="!{action}:{scope}"
        )
        release_data = ReleaseDataTree(semantic_commits)
        version = Version(major=10, minor=3, patch=0)

        markup = _render_release(version, release_data=release_data)

        assert markup == (
            f"10.3.0 ({str(date.today())})\n"
            "-------------------\n\n"
            "Bug Fixes\n"
            "~~~~~~~~~\n\n"
            "* thing - Fix a thing\n\n"
        )

    def test_release_with_features(self, commit_registry):
        reg = commit_registry
        version = Version(major=10, minor=3, patch=0)
        semantic_commits = commit_analyzer(
            [reg["ccaa185"], reg["bc0bcab"], reg["a6b655f"]],
            label_pattern="!{action}:{scope}",
        )
        release_data = ReleaseDataTree(semantic_commits)
        markup = _render_release(version, release_data=release_data)

        assert markup == (
            f"10.3.0 ({str(date.today())})\n"
            "-------------------\n\n"
            "Features\n"
            "~~~~~~~~\n\n"
            "* Add a thing\n"
            "* music - Add more music please\n"
            "* cli - Add a cli tool\n\n"
        )


@parametrize(
    "file_content, expected",
    [
        ("", "New Content\n"),
        (
            ("Changelog\n" "=========\n" "\n" "1.0.0\n" "~~~~~\n" "Content\n"),
            (
                "Changelog\n"
                "=========\n"
                "\n"
                "New Content\n"
                "1.0.0\n"
                "~~~~~\n"
                "Content\n"
            ),
        ),
        (
            ("One Paragraph.\n" "Two Paragraphs.\n" "\n" "Another Paragraph\n"),
            (
                "One Paragraph.\n"
                "Two Paragraphs.\n"
                "\n"
                "Another Paragraph\n"
                "New Content\n"
            ),
        ),
    ],
    ids=["Empty file", "File with a release section", "File without release section"],
)
@patch("braulio.files._render_release", return_value="New Content\n")
def test_update_chglog(
    mock_render_release, isolated_filesystem, file_content, expected
):
    cur_version = Version("1.0.0")
    new_version = Version("2.0.0")

    with isolated_filesystem:
        path = Path.cwd() / "CHANGELOG.rst"
        path.write_text(file_content)

        update_chglog(path, cur_version, new_version, {})

        mock_render_release.assert_called_with(new_version, {})

        text = path.read_text()

        assert text == expected


class Test_update_files:
    def test_files_missing_version_string(self, fake_repository):
        paths = ["setup.py", "black/__init__.py"]

        with fake_repository("black"):
            message = 'Unable to find a version string to update in "setup.py"'

            with pytest.raises(ValueError, match=message):
                update_files(paths, "4.0.0", "4.1.0")

    def test_file_update(self, fake_repository):
        paths = ["setup.py", "black/__init__.py"]

        with fake_repository("black"):

            update_files(paths, "4.1.3", "5.0.0")

            __init__file = (Path.cwd() / "black" / "__init__.py").read_text()
            setup_file = (Path.cwd() / "setup.py").read_text()

        assert setup_file == (
            "from setuptools import setup\n"
            "\n"
            "setup(\n"
            "    name='black',\n"
            "    author='Mr. Black',\n"
            "    description='Jus an example',\n"
            "    version='5.0.0',\n"
            "    zip_safe=False,\n"
            ")\n"
        )

        assert __init__file == (
            "'''\n"
            "    Black project\n"
            "    ~~~~~~~~~~~~~\n"
            "'''\n"
            "\n"
            "__author__ = 'Mr. Black'\n"
            "__email__ = 'black@example.test'\n"
            "__version__ = '5.0.0'\n"
            "\n"
            "\n"
            "def example():\n"
            "    pass\n"
        )



import re
import click
from pathlib import Path
from click import style
from braulio.git import Git, commit_analyzer
from braulio.version import Version, get_next_version, VERSION_STRING_PATTERN
from braulio.config import Config, update_config_file
from braulio.files import (
    find_chglog_file,
    create_chglog_file,
    update_chglog,
    update_files,
    ReleaseDataTree,
    DEFAULT_CHANGELOG,
)


prefix_mark = style(" › ", fg="blue", bold=True)
check_mark = click.style("✓", fg="green", bold=True)
x_mark = click.style("✗", fg="red", bold=True)


def msg(message, prefix=prefix_mark, suffix="", silence=False, nl=True):

    if not silence:
        click.echo(f"{prefix}{message}{suffix}", nl=nl)


def label(text):
    spaces = " " * (17 - len(text))
    return style(f"{text}{spaces}:", fg="blue", bold=True)


@click.group()
@click.version_option(prog_name="Braulio")
@click.pass_context
def cli(ctx):

    config = Config()
    ctx.obj = config

    try:
        Version.set_stages(config.stages.values())
    except ValueError as e:
        ctx.fail(e)

    ctx.default_map = {
        "release": {
            "tag_flag": config.tag,
            "commit_flag": config.commit,
            "confirm_flag": config.confirm,
            "label_position": config.label_position,
            "label_pattern": config.label_pattern,
            "tag_pattern": config.tag_pattern,
            "current_version": config.current_version,
            "changelog_file": config.changelog_file,
            "message": config.message,
        }
    }


@cli.command()
@click.option(
    "--changelog-file",
    "changelog_name",
    help="A name for the changelog file to be created",
)
def init(changelog_name):
    """Setup your project."""

    changelog_path = find_chglog_file()
    create_changelog_flag = True
    mark = style("?", fg="blue", bold=True)

    if not changelog_name:
        if changelog_path:
            filename = style(changelog_path.name, fg="blue", bold=True)
            message = f" {mark} {filename} was found." " Is this the changelog file?"

            if click.confirm(message):
                changelog_name = changelog_path.name
                create_changelog_flag = False

        if create_changelog_flag:
            message = f" {mark} Enter a name for the changelog:"
            changelog_name = click.prompt(message, default=DEFAULT_CHANGELOG)

    if create_changelog_flag:
        create_chglog_file(changelog_name)

    if changelog_name and create_changelog_flag:
        update_config_file("changelog_file", changelog_name)


def bump_option_validator(ctx, param, value):
    """In case a value is provided checks that it is a valid version string. If
    is not thrown :class:`click.UsageError`.

    Return a :class:`~braulio.version.Version` object or **None**.
    """

    if value:
        try:
            value = Version(value)
        except ValueError:
            ctx.fail(f"{x_mark} {value} is not a valid version string")

    return value


def files_callback(ctx, param, value):
    if value:
        return value

    config = ctx.obj
    return config.files


def changelog_file_option_validator(ctx, param, value):
    """Checks that the given file path exists in the current working directory.

    Returns a :class:`~pathlib.Path` object. If the file does not exist raises
    a :class:`~click.UsageError` exception.
    """

    path = Path(value)

    if not path.exists():
        filename = click.style(path.name, fg="blue", bold=True)
        ctx.fail(
            "\n"
            f" {x_mark} Unable to find {filename}\n"
            '   Run "$ brau init" to create one'
        )

    return path


def message_option_validator(ctx, param, value):
    """A commit template must have the **{new_version}** placeholder."""

    if not value or "{new_version}" not in value:
        ctx.fail("Missing {new_version} placeholder in " f"{value}.")

    return value


def tag_pattern_option_validator(ctx, param, value):
    """The provided string must contain **{version}** placeholder in order to
    be valid. Otherwise :class:`click.UsageError` is raised.
    """

    if not value or "{version}" not in value:
        ctx.fail("Missing {version} placeholder in tag_pattern.")

    return value


def current_version_option_validator(ctx, param, value):
    """If a version string is provided, validates it. Otherwise it tries
    to determine the current version from the last Git tag that matches
    ``tag_pattern`` option.

    Return a :class:`~braulio.version.Version` object or **None**.
    """

    current_version = None

    if value:
        try:
            current_version = Version(value)
        except ValueError:
            ctx.fail(f"{value} is not a valid version string")

    # Look for the last git tag for the curren version
    git = Git()
    tag_pattern = ctx.params["tag_pattern"]

    if current_version:
        tag_pattern = re.escape(tag_pattern.format(version=current_version.string))
    else:
        tag_pattern = re.escape(tag_pattern).replace(
            r"\{version\}", VERSION_STRING_PATTERN
        )

    tag_regex = re.compile(tag_pattern)

    for tag in git.tags:
        match = tag_regex.match(tag.name)

        if match:
            ctx.params["current_tag"] = tag

            if not current_version:
                current_version = Version(**match.groupdict(default=0))

            break

    return current_version


def label_pattern_option_validator(ctx, param, value):
    """Checks that a given string has all the required placeholders. The
    possible placeholders are **{action}**, **{scope}** and **{subject}**.

    ``{action}`` are always required, ``{scope}`` is optional and
    ``{subject}`` are required only when ``label_position`` option is set to
    ``header``.

    If the pattern is invalid, raises :class:`click.UsageError`.
    """

    missings = []
    label_position = ctx.params["label_position"]

    if "{action}" not in value:
        missings.append(style("{action}", fg="red", bold=True))

    if label_position == "header" and "{subject}" not in value:
        missings.append(style("{subject}", fg="red", bold=True))

    if missings:
        message = "\n"

        for placeholder in missings:
            message += f"  {x_mark} Missing {placeholder} placeholder.\n"

        ctx.fail(message)

    return value


@cli.command()
@click.option("--major", "bump_type", flag_value="major", help="Major version bump.")
@click.option("--minor", "bump_type", flag_value="minor", help="Minor version bump.")
@click.option("--patch", "bump_type", flag_value="patch", help="Patch version bump.")
@click.option(
    "--bump",
    callback=bump_option_validator,
    help="Bump to a given version arbitrarily.",
)
@click.option(
    "--commit/--no-commit",
    "commit_flag",
    default=True,
    help="Enable/disable release commit",
)
@click.option(
    "--message", callback=message_option_validator, help="Customizes commit message."
)
@click.option(
    "--tag/--no-tag", "tag_flag", default=True, help="Enable/disable version tagging."
)
@click.option(
    "--changelog-file",
    "changelog_file",
    type=click.Path(),
    callback=changelog_file_option_validator,
    help="Specify the changelog file.",
)
@click.option(
    "--label-position",
    type=click.Choice(["header", "footer"]),
    is_eager=True,
    help="Where the label is located in the commit message.",
)
@click.option(
    "--label-pattern",
    callback=label_pattern_option_validator,
    help="Pattern to identify labels in commit messages.",
)
@click.option(
    "--tag-pattern",
    callback=tag_pattern_option_validator,
    is_eager=True,
    help="Pattern for Git tags that represent versions",
)
@click.option(
    "--current-version",
    help="Manually specify the curren version.",
    callback=current_version_option_validator,
)
@click.option(
    "-y", "confirm_flag", is_flag=True, default=False, help="Don't ask for confirmation"
)
@click.argument(
    "files", nargs=-1, type=click.Path(exists=True), callback=files_callback
)
@click.pass_context
def release(
    ctx,
    bump,
    bump_type,
    commit_flag,
    message,
    tag_flag,
    confirm_flag,
    changelog_file,
    files,
    label_pattern,
    label_position,
    tag_pattern,
    current_version,
    current_tag=None,
):

    """Release a new version.

    Determines the next version by inspecting commit messages, updates the
    changelog, commit the changes and tag the repository with the new version.
    """

    git = Git()
    current_tag_name = current_tag.name if current_tag else None
    commit_list = git.log(_from=current_tag_name)

    msg(f'{label("Current version")} {current_version}')
    msg(f'{label("Commits found")} {len(commit_list)} since last release')

    if not commit_list:
        click.echo(" › Nothing to release.")
        ctx.exit()

    semantic_commits = commit_analyzer(commit_list, label_pattern, label_position)

    release_data = ReleaseDataTree(semantic_commits)

    # --bump must have precedence over any of --major, --minor or --patch
    bump_version_to = bump.string if bump else bump_type

    # Any manual bump must have precedence over bump part determined from
    # commit messages
    bump_version_to = bump_version_to or release_data.bump_version_to

    # If there isn't a current version, assume version 0.0.0
    current_version = current_version or Version()

    if current_version.stage != "final" and not bump:
        new_version = current_version.bump("final")
    else:
        new_version = get_next_version(current_version, bump_version_to)

    if not new_version:
        msg("The release of a lower versions is not supported for now.")
        ctx.abort()

    new_tag_name = tag_pattern.format(version=new_version.string)

    msg(f'{label("New version")} {new_version}')
    msg(f'{label("Changelog file")} {changelog_file.name}')

    # Messages about what tasks will be performed
    msg("Braulio will perform the next tasks :")
    msg(f"        Update {len(files) + 1} files.", prefix="")
    msg("        Add a release commit.", prefix="", silence=not commit_flag)
    msg(
        f"        Tag the repository with {new_tag_name}",
        prefix="",
        silence=not tag_flag,
    )

    msg("", prefix="")  # Print just a new line

    if confirm_flag or click.confirm(f"{prefix_mark}Continue?"):

        msg("Update changelog ", nl=False)

        update_chglog(
            changelog_file,
            new_version=new_version,
            current_version=current_version,
            release_data=release_data,
        )

        msg(check_mark, prefix="")

        try:
            update_files(files, str(current_version), str(new_version))
        except ValueError as e:
            click.echo(e)
            ctx.abort()

        if commit_flag:
            message_args = {"new_version": new_version.string}

            if "{current_version}" in message:
                message_args["current_version"] = current_version.string

            commit_message = message.format(**message_args)

            msg(f"Add commit: {commit_message}", nl=False)

            files = [str(changelog_file)] + list(files)
            git.commit(commit_message, files=files)
            msg(f" {check_mark}", prefix="")

        if tag_flag:
            msg(f"Add tag {new_tag_name}", nl=False)
            git.tag(new_tag_name)
            msg(f" {check_mark}", prefix="")

        if "current_version" in ctx.obj.cfg_file_options:
            update_config_file("current_version", new_version.string)

        msg(f"Version {new_version} released successfully", suffix=" 🎉")



import re
from pathlib import Path
from typing import NamedTuple
from subprocess import run, PIPE, CalledProcessError

hash_pattern = re.compile("(?<=commit )\w{40}$", re.M)


def _run_command(command):
    captured = run(command, stdout=PIPE, stderr=PIPE, check=True)
    return captured.stdout.decode()


def _run_git_tag_command():
    command = [
        "git",
        "tag",
        "--sort=creatordate",
        "--format=%(creatordate:short)%09%(refname:strip=2)",
    ]
    return _run_command(command)


class Tag:
    def __init__(self, text):
        self.text = text
        self.date = text[:10]
        self.name = text[10:].strip()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f'Tag("{self.text}")'


class Commit:
    def __init__(self, text):
        self.text = text
        lines = text.strip().split("\n")

        # Commit hash
        self.hash = lines[0][7:]

        # Commit message
        msg_lines = [line[4:] for line in lines[4:]]
        self.message = "\n".join(msg_lines)

        # Commit message header
        self.header = msg_lines[0]
        self.footer = msg_lines[-1]

        # Commit message body
        self.body = None

        if len(msg_lines) > 2 and msg_lines[1] == "":
            self.body = "\n".join(msg_lines[2:])

    def __repr__(self):
        return f"Commit('{self.header}')"


patter = re.compile("^commit (?:.+?(?=commit \w{40})|.+$)", flags=re.M | re.S)


def _extract_commit_texts(git_log_text):
    return re.findall(patter, git_log_text)


class Git:
    def add(self, *files):
        """Add one or more files to the index running git-add."""

        try:
            _run_command(("git", "add") + files)
        except CalledProcessError:
            # Only if the command fails we check if the files
            # exist, because git-add most of the time fails when
            # the provided files are not found.
            for f in files:
                if not Path(f).exists():
                    raise FileNotFoundError(f"No such file or directory: {f}")

    def commit(self, message, files=None):
        """Run git-commit."""

        if files:
            self.add(*files)

        return _run_command(["git", "commit", "-m", f'"{message}"'])

    def log(self, _from=None, to=None):
        """Run git-log."""

        command = ["git", "log"]

        if _from:
            to = "HEAD" if not to else to
            revision_range = f"{_from}..{to}"
            command.append(revision_range)

        git_log_text = _run_command(command)
        commit_text_lst = _extract_commit_texts(git_log_text)

        return [Commit(commit_text) for commit_text in commit_text_lst]

    def tag(self, name=None):
        """Create and list tag objects running git-tag command"""

        command = ["git", "tag"]

        if not name:
            command.extend(
                [
                    "-l",
                    "--sort=creatordate",
                    "--format=%(creatordate:short)%09%(refname:strip=2)",
                ]
            )

            command_output = _run_command(command).strip()

            if command_output == "":
                return []

            tag_text_list = command_output.split("\n")
            tag_list = [Tag(text) for text in tag_text_list]

            return list(reversed(tag_list))

        command.extend(["-a", name, "-m", '""'])
        return _run_command(command)

    @property
    def tags(self):
        if not hasattr(self, "_tag_list"):
            self._tag_list = self.tag()
        return self._tag_list


class SemanticCommit(NamedTuple):
    subject: str
    message: str
    action: str
    scope: str


def commit_analyzer(commits, label_pattern, label_position="footer"):
    """Analyzes a list of :class:`~braulio.git.Commit` objects searching for
    messages that match a given message convention and extract metadata from
    them.

    A message convention is determined by ``label_pattern``, which is not a
    regular expression pattern. Instead it must be a string literals with
    placeholders that indicates metadata information in a given position of
    the commit message. The possible placeholders are ``{action}``, ``{scope}``
    and ``{subject}``.

    The ``label_position`` argument dictates where (header|footer) to look in
    the commit message for the pattern passed in ```label_pattern``.

    ``{subject}`` must be included in ``label_pattern`` just if the metadata is
    in the header, otherwise must be omitted.

    Examples.

    If ``label_position`` is equal to **header**, in order to match the next
    commit message::

        fix(cli): Ensure --help option doesn't hang

    The pattern must be ``{action}({scope}): {subject}``, where the metadata
    information extracted will be::

        {
            'action': 'fix',
            'scope': 'cli',
            'subject': 'Ensure --help option doesn't hang'
        }
    """

    # Internally, a real regular expression pattern is used
    pattern_string = re.escape(label_pattern)

    # Capturing group patterns
    action_cgp = r"(?P<action>\w+)"
    scope_cgp = r"(?P<scope>\w*)"
    subject_cgp = r"(?P<subject>.+)"

    pattern_string = pattern_string.replace(r"\{action\}", action_cgp).replace(
        r"\{scope\}", scope_cgp
    )
    if label_position == "header":
        pattern_string = pattern_string.replace(r"\{subject\}", subject_cgp)

    regexp_pattern = re.compile(pattern_string)

    semantic_commits = []

    for commit in commits:
        text = commit.header if label_position == "header" else commit.footer
        match = regexp_pattern.search(text)

        if not match:
            continue

        metadata = match.groupdict()

        if label_position == "footer":
            metadata["subject"] = commit.header

        sc = SemanticCommit(
            subject=metadata["subject"].strip(),
            action=metadata["action"],
            scope=metadata.get("scope") or None,
            message=commit.message,
        )

        semantic_commits.append(sc)

    return semantic_commits



import re
import click
from collections import UserDict
from datetime import date
from pathlib import Path


KNOWN_CHANGELOG_FILES = ("HISTORY.rst", "CHANGELOG.rst", "CHANGES.rst")
DEFAULT_CHANGELOG = KNOWN_CHANGELOG_FILES[0]


class ReleaseDataTree(UserDict):
    """Takes a list of :class:`~braulio.git.SemanticCommit` objects, and
    classify them by type of action (feat, fix, refactor, etc) and then by
    the scope of the commit.

    This is a dict like class, so, the found actions can be accessed normally
    by the key name.

    Commits without an specified action are filter out.
    """

    def __init__(self, semantic_commits):

        self.data = {}

        # If this release is braking thing
        is_breaking = False

        for commit in semantic_commits:
            if commit.action:

                is_breaking = is_breaking or "BREAKING CHANGE" in commit.message

                action, scope = commit.action, commit.scope

                if action not in self:
                    self[action] = {"scopeless": []}

                if not scope:
                    self[action]["scopeless"].append(commit)
                    continue
                else:
                    if scope not in self[action]:
                        self[action][commit.scope] = []

                    self[action][scope].append(commit)

        bump_type = "major" if is_breaking else "minor" if "feat" in self else "patch"

        self._bump_version_to = bump_type

    @property
    def bump_version_to(self):
        return self._bump_version_to


def find_chglog_file():
    for file_name in KNOWN_CHANGELOG_FILES:
        path = Path.cwd() / file_name

        if path.is_file():
            return path
    return None


def create_chglog_file(name=None):
    file_name = name or DEFAULT_CHANGELOG
    path = Path.cwd() / file_name
    path.touch()
    path.write_text(_render_title("History"))

    mark = click.style("✓", fg="green")
    click.echo(f" {mark} {file_name} created succesfully.")


def _render_title(title, level=1):
    underlines = ["=", "-", "~"]
    underline = underlines[level - 1] * len(title)
    return f"{title}\n{underline}\n\n"


def _render_subtitle(commits):
    markup = ""

    for commit in commits:
        markup += f"  - {commit.subject}\n"

    return markup


def _render_list(scope_dict):
    markup = ""

    for commit in scope_dict["scopeless"]:
        markup += f"* {commit.subject}\n"

    del scope_dict["scopeless"]

    for scope_name, commit_lst in scope_dict.items():
        markup += f"* {scope_name}"

        if len(commit_lst) > 1:
            markup += "\n\n" + _render_subtitle(commit_lst)
        else:
            markup += f" - {commit_lst[0].subject}\n"

    return markup + "\n"


def _render_release(version, release_data):
    today = str(date.today())
    title = f"{version.string} ({today})"
    markup = _render_title(title, level=2)

    if "fix" in release_data:
        markup += _render_title("Bug Fixes", level=3)
        markup += _render_list(release_data["fix"])

    if "feat" in release_data:
        markup += _render_title("Features", level=3)
        markup += _render_list(release_data["feat"])

    return markup


title_adornment_pattern = re.compile("^(?:=|~|-|\*|`|')+$")


def is_title(first, second, third):

    if not first or not second:
        return False

    if (
        title_adornment_pattern.match(first)
        and len(first) == len(second)
        and first == third
    ):
        return True

    if first == "\n" and third:
        first, second = second, third

    if len(first) == len(second) and title_adornment_pattern.match(second):
        return True

    return False


def _split_chglog(path, title):
    """Split a RST file text in two parts. The title argument determine the
    split point. The given title goes in the bottom part. If the title is not
    found everything goes in the top part.

    Return a tuple with the top and bottom parts.
    """

    with path.open() as f:
        doc = f.readlines()

    has_title = False

    for idx, curr_line in enumerate(doc):

        if title in curr_line:
            prev_line = doc[idx - 1] if idx - 1 < len(doc) else "\n"
            next_line = doc[idx + 1] if idx + 1 < len(doc) else None

            if is_title(prev_line, curr_line, next_line):
                idx = idx if prev_line == "\n" else idx - 1
                has_title = True
                break

    if has_title:
        top, bottom = doc[:idx], doc[idx:]
    else:
        top, bottom = doc, []

    return "".join(top), "".join(bottom)


def update_chglog(path, current_version, new_version, release_data):
    top, bottom = _split_chglog(path, title=current_version.string)
    markup = _render_release(new_version, release_data)

    path.write_text(top + markup + bottom)


version_pattern = re.compile("_?_?version_?_?\s?=\s?(?:'|\")")


def update_files(paths, current_version, new_version):
    path_list = [Path(p) for p in paths]

    for path in path_list:
        if not path.is_file():
            click.echo(f"The file {path} is invalid or does not exist")

        text = ""
        string_found = False

        with path.open() as f:
            for line in f:
                if version_pattern.search(line) and current_version in line:
                    line = line.replace(current_version, new_version)
                    string_found = True

                text += line

        if not string_found:
            raise ValueError(f'Unable to find a version string to update in "{path}"')

        path.write_text(text)



import pytest
from configparser import ConfigParser
from unittest.mock import patch
from pathlib import Path
from click.testing import CliRunner
from braulio.cli import cli
from braulio.files import DEFAULT_CHANGELOG, KNOWN_CHANGELOG_FILES


parametrize = pytest.mark.parametrize


@pytest.fixture(params=KNOWN_CHANGELOG_FILES)
def known_changelog_file(request):
    return request.param


@parametrize("_input", ["CUSTOM.rst", None])
def test_empty_repository(_input):
    runner = CliRunner()

    with runner.isolated_filesystem():
        result = runner.invoke(cli, ["init"], input=_input)

        assert result.exit_code == 0
        assert "Enter a name for the changelog" in result.output

        filename = DEFAULT_CHANGELOG if not _input else _input
        changelog_path = Path.cwd() / filename

        assert changelog_path.exists()


def test_changelog_filename_option():
    runner = CliRunner()

    with runner.isolated_filesystem():
        command = ["init", "--changelog-file=CUSTOM.rst"]
        result = runner.invoke(cli, command, input="y")

        changelog_path = Path.cwd() / "CUSTOM.rst"
        config_path = Path.cwd() / "setup.cfg"

        assert result.exit_code == 0, result.exception
        assert "CUSTOM.rst created succesfully." in result.output
        assert "Updated setup.cfg changelog_file option" in result.output

        assert changelog_path.exists()
        assert config_path.exists()

        config = ConfigParser()
        config.read("setup.cfg")

        assert config.has_section("braulio")
        assert config.get("braulio", "changelog_file") == "CUSTOM.rst"


@pytest.mark.parametrize("_input", ["y", "n"])
@patch("braulio.cli.create_chglog_file", autospec=True)
def test_known_changelog_found(
    mock_create_changelog_file, known_changelog_file, _input
):

    runner = CliRunner()

    with runner.isolated_filesystem():
        file_path = Path.cwd() / known_changelog_file
        file_path.touch()

        result = runner.invoke(cli, ["init"], input=_input)

        assert f"{known_changelog_file} was found" in result.output
        assert "Is this the changelog file?" in result.output

        if _input == "y":
            mock_create_changelog_file.assert_not_called()
        else:
            mock_create_changelog_file.assert_called()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['Click']

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    author="José María Domínguez Moreno",
    author_email='miso.0b11@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
    ],
    description="A command line tool to handle changelogs using Git commit messages",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='braulio',
    name='braulio',
    packages=find_packages(include=['braulio']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/mbarakaja/braulio',
    version='0.3.0b0',
    zip_safe=False,
    entry_points='''
        [console_scripts]
        brau=braulio.cli:cli
    ''',
)



import pytest
from unittest.mock import patch as mock_patch
from collections import OrderedDict
from braulio.version import (
    validate_version_str,
    Version,
    get_next_version,
    parse_version_string_parts,
)


parametrize = pytest.mark.parametrize


version_strings = [
    # ---- invalids -----
    ("abc", None),
    ("a.b.c", None),
    ("1.1.1.1", None),
    ("1.", None),
    ("1.1.", None),
    ("1.1.1.", None),
    ("1.2.3.dev", None),
    ("1.2.3a", None),
    ("1.2.3alpha", None),
    ("1.2.3beta", None),
    ("1.2.3b3ta0", None),
    ("1.2.3rc", None),
    # ---- valids -----
    ("1", {"major": 1, "minor": 0, "patch": 0, "stage": None, "n": 0}),
    ("1.1", {"major": 1, "minor": 1, "patch": 0, "stage": None, "n": 0}),
    ("1.1.1", {"major": 1, "minor": 1, "patch": 1, "stage": None, "n": 0}),
    ("0", {"major": 0, "minor": 0, "patch": 0, "stage": None, "n": 0}),
    ("0.0", {"major": 0, "minor": 0, "patch": 0, "stage": None, "n": 0}),
    ("0.0.0", {"major": 0, "minor": 0, "patch": 0, "stage": None, "n": 0}),
    ("7.10.3", {"major": 7, "minor": 10, "patch": 3, "stage": None, "n": 0}),
    ("30.1.11", {"major": 30, "minor": 1, "patch": 11, "stage": None, "n": 0}),
    ("010.05.04", {"major": 10, "minor": 5, "patch": 4, "stage": None, "n": 0}),
    ("1.2.3.dev1", {"major": 1, "minor": 2, "patch": 3, "stage": "dev", "n": 1}),
    ("1.2.3a3", {"major": 1, "minor": 2, "patch": 3, "stage": "a", "n": 3}),
    ("1.2.3alpha5", {"major": 1, "minor": 2, "patch": 3, "stage": "alpha", "n": 5}),
]


@parametrize("string, expected", version_strings)
def test_validate_version_str(string, expected):
    assert validate_version_str(string) is bool(expected)


@parametrize("string, parts", version_strings)
def test_parse_version_string_parts(string, parts):
    result = parse_version_string_parts(string)
    assert result == parts


class TestVersion:
    def test_default(self):
        version = Version()

        assert version.major == 0
        assert version.minor == 0
        assert version.patch == 0
        assert version.string == "0.0.0"
        assert version.stage == "final"
        assert version.n is 0

    def test_invalid_version_string_argument(self):
        with pytest.raises(ValueError):
            Version("invalid")

    @parametrize(
        "string, major, minor, patch, stage, n",
        [
            ("1.2.3", 1, 2, 3, "final", 0),
            ("1", 1, 0, 0, "final", 0),
            ("2.1a2", 2, 1, 0, "a", 2),
            ("3.1.4.dev0", 3, 1, 4, "dev", 0),
        ],
    )
    def test_deserializing(self, string, minor, major, patch, stage, n):
        stages = {
            "dev": "{major}.{minor}.{patch}dev{n}",
            "a": "{major}.{minor}.{patch}a{n}",
            "final": "{major}.{minor}.{patch}",
        }

        with mock_patch.object(Version, "_stages", stages):
            version = Version(string)

            assert version.major == major
            assert version.minor == minor
            assert version.patch == patch
            assert version.stage == stage
            assert version.n == n

    @parametrize(
        "major, minor, patch, expected", [(1, 2, 3, "1.2.3"), (1, 0, 0, "1.0.0")]
    )
    def test_default_serializers(self, major, minor, patch, expected):
        version = Version(major=major, minor=minor, patch=patch)

        assert version.string == expected

    @parametrize(
        "major, minor, patch, stage, n, expected",
        [
            (1, 2, 3, None, 0, "1.2.3"),
            (1, 0, 0, None, 0, "1.0.0"),
            (2, 1, 0, "alpha", 2, "2.1.0alpha2"),
            (2, 1, 0, "beta", 1, "2.1.0beta1"),
            (8, 0, 0, "gamma", 4, "8.0.0gamma4"),
        ],
    )
    def test_custom_serializers(self, major, minor, patch, stage, n, expected):
        stages = OrderedDict(
            alpha="{major}.{minor}.{patch}alpha{n}",
            beta="{major}.{minor}.{patch}beta{n}",
            gamma="{major}.{minor}.{patch}gamma{n}",
            final="{major}.{minor}.{patch}",
        )

        with mock_patch.object(Version, "_stages", stages):
            version = Version(major=major, minor=minor, patch=patch, stage=stage, n=n)

            assert version.string == expected

    def test_unknown_stage(self):
        error = "gama is an unknown stage"

        with pytest.raises(ValueError, match=error):
            Version("2.0.7gama0")

    def test_set_stages(self):
        stages = [
            "{major}.{minor}.{patch}a{n}",
            "{major}.{minor}.{patch}beta{n}",
            "{major}.{minor}.{patch}gamma{n}",
            "{major}.{minor}.{patch}",
        ]

        Version.set_stages(stages)

        assert Version._stages == OrderedDict(
            a="{major}.{minor}.{patch}a{n}",
            beta="{major}.{minor}.{patch}beta{n}",
            gamma="{major}.{minor}.{patch}gamma{n}",
            final="{major}.{minor}.{patch}",
        )

    def test_set_stages_with_invalid_serializer(self):
        stages = [
            "{major}.{minor}.{patch}a{n}",
            "{major}.{patch}beta{n}",
            "{major}.{minor}.{patch}gamma{n}",
            "{major}.{minor}.{patch}",
        ]

        message = "{major}.{patch}beta{n} is an invalid serializer"

        with pytest.raises(ValueError, match=message):
            Version.set_stages(stages)

    @parametrize(
        "current, bump_part, expected",
        [
            ("1.2.3", "major", "2.0.0"),
            ("1.2.3", "minor", "1.3.0"),
            ("1.2.3", "patch", "1.2.4"),
            ("1.2.3.dev0", "dev", "1.2.3.dev1"),
            ("1.2.3.dev5", "beta", "1.2.3beta0"),
            ("1.2.3.dev5", "final", "1.2.3"),
            ("1.2.3beta0", "final", "1.2.3"),
        ],
        ids=[
            "major",
            "minor",
            "patch",
            "stage numerical component",
            "to next stage",
            "from pre-release stage to final stage",
            "from first stage to final stage",
        ],
    )
    def test_bump(self, current, bump_part, expected):
        stages = OrderedDict(
            dev="{major}.{minor}.{patch}.dev{n}",
            beta="{major}.{minor}.{patch}beta{n}",
            final="{major}.{minor}.{patch}",
        )

        with mock_patch.object(Version, "_stages", stages):
            version = Version(current)
            new_version = version.bump(bump_part)

            assert new_version == Version(expected)
            assert new_version is not version

    @pytest.mark.wip
    @parametrize(
        "current, bump_part, message",
        [
            ("1.2.3beta0", "dev", "dev stage is previous to 1.2.3beta0"),
            ("1.2.3beta0", "major", "Can't do a major version bump"),
            ("1.2.3beta0", "minor", "Can't do a minor version bump"),
            ("1.2.3beta0", "patch", "Can't do a patch version bump"),
            ("1.2.3", "final", "1.2.3 is already in final stage"),
            ("1.2.3", "unknown", "Unknown unknown stage"),
        ],
        ids=[
            "to previous stage",
            "major part in pre-release stage",
            "minor part in pre-release stage",
            "patch part in pre-release stage",
            "from final to final stage",
            "unknown stage",
        ],
    )
    def test_invalid_bump(self, current, bump_part, message):
        stages = OrderedDict(
            dev="{major}.{minor}.{patch}.dev{n}",
            beta="{major}.{minor}.{patch}beta{n}",
            final="{major}.{minor}.{patch}",
        )

        with mock_patch.object(Version, "_stages", stages):
            with pytest.raises(ValueError, match=message):
                version = Version(current)
                version.bump(bump_part)

    @parametrize(
        "left, right",
        [
            (Version(), 1),
            (Version(), "2.23.8"),
            (Version(), None),
            (Version(), True),
            (1, Version()),
            ("2.23.8", Version()),
            (None, Version()),
            (True, Version()),
        ],
    )
    def test_comparators_with_invalid_values(self, left, right):
        error_message = "not supported"

        with pytest.raises(TypeError, match=error_message):
            left > right

        with pytest.raises(TypeError, match=error_message):
            left < right

        with pytest.raises(TypeError, match=error_message):
            left <= right

        with pytest.raises(TypeError, match=error_message):
            left >= right

        with pytest.raises(TypeError, match=error_message):
            left == right

        with pytest.raises(TypeError, match=error_message):
            left != right

    def test_repr(self):
        assert repr(Version("3.2.1")) == "Version('3.2.1')"

    def test_str(self):
        assert str(Version("3.2.1")) == "3.2.1"


# This tuple defines the expected result of comparing
# two Version instances using <, >, ==, <= operators.
version_strings_to_compare = (
    ("4.0.0", "4.0.0", False, False, True),
    ("2.0.1", "2.0.0", False, True, False),
    ("1.4.0", "1.3.0", False, True, False),
    ("3.0.0", "2.23.8", False, True, False),
    ("2.0.0", "2.0.1", True, False, False),
    ("1.3.0", "1.4.0", True, False, False),
    ("2.23.8", "3.0.0", True, False, False),
    ("1.0.0.dev1", "1.0.0.dev0", False, True, False),
    ("0.1.0.dev1", "0.1.0.dev0", False, True, False),
    ("0.0.1.dev1", "0.0.1.dev0", False, True, False),
    ("2.0.0.dev0", "2.0.0.dev0", False, False, True),
    ("0.2.0.dev0", "0.2.0.dev0", False, False, True),
    ("0.0.2.dev0", "0.0.2.dev0", False, False, True),
    ("3.0.0beta0", "3.0.0.dev7", False, True, False),
    ("0.3.0beta0", "0.3.0.dev7", False, True, False),
    ("0.0.3beta0", "0.0.3.dev7", False, True, False),
    ("5.0.0.dev0", "4.0.0beta1", False, True, False),
    ("0.5.0.dev0", "0.4.0beta1", False, True, False),
    ("0.0.5.dev0", "0.0.4beta1", False, True, False),
    ("6.0.0", "6.0.0beta1", False, True, False),
    ("0.6.0", "0.6.0beta1", False, True, False),
    ("0.0.6", "0.0.6beta1", False, True, False),
    ("1.0.0.dev0", "1.0.0.dev1", True, False, False),
    ("0.1.0.dev0", "0.1.0.dev1", True, False, False),
    ("0.0.1.dev0", "0.0.1.dev1", True, False, False),
    ("3.0.0dev3", "3.0.0beta0", True, False, False),
    ("0.3.0dev3", "0.3.0beta0", True, False, False),
    ("0.0.3dev3", "0.0.3beta0", True, False, False),
    ("6.0.0beta5", "6.0.0", True, False, False),
    ("0.6.0beta5", "0.6.0", True, False, False),
    ("0.0.6beta5", "0.0.6", True, False, False),
)


@parametrize("left, right, is_less, is_greater, is_equal", version_strings_to_compare)
def test_comparison_operators(left, right, is_less, is_greater, is_equal):
    stages = OrderedDict(
        dev="{major}.{minor}.{patch}.dev{n}",
        beta="{major}.{minor}.{patch}beta{n}",
        final="{major}.{minor}.{patch}",
    )

    with mock_patch.object(Version, "_stages", stages):
        assert (Version(left) < Version(right)) is is_less
        assert (Version(left) > Version(right)) is is_greater
        assert (Version(left) == Version(right)) is is_equal
        assert (Version(left) <= Version(right)) is (is_less or is_equal)
        assert (Version(left) >= Version(right)) is (is_greater or is_equal)
        assert (Version(left) != Version(right)) is not is_equal


@parametrize(
    "current_version, bump_version_to, expected",
    [
        (None, "0.0.0", "0.0.0"),
        (None, "7.0.0", "7.0.0"),
        (None, "0.8.0", "0.8.0"),
        (None, "0.0.12", "0.0.12"),
        (None, "major", "1.0.0"),
        (None, "minor", "0.1.0"),
        (None, "patch", "0.0.1"),
        (None, "patch", "0.0.1"),
        ("0.0.0", "0.0.0", None),
        ("0.0.1", "0.0.0", None),
        ("0.0.1", "0.0.2", "0.0.2"),
        ("4.4.7", "major", "5.0.0"),
        ("4.4.7", "minor", "4.5.0"),
        ("4.4.7", "patch", "4.4.8"),
    ],
)
def test_get_next_version(current_version, bump_version_to, expected):
    current_version = Version(current_version) if current_version else None
    expected = Version(expected) if expected else None

    new_version = get_next_version(current_version, bump_version_to)

    assert new_version == expected



import pytest
from collections import namedtuple
from subprocess import CalledProcessError, PIPE
from unittest.mock import patch
from braulio.git import _run_command, Git, Commit, Tag, commit_analyzer


parametrize = pytest.mark.parametrize


class TestRunCommand:
    @patch("braulio.git.run")
    def test_call_to_subprocess_run(self, mocked_run):

        output = _run_command(["command", "subcomand"])

        mocked_run.assert_called_with(
            ["command", "subcomand"], stdout=PIPE, stderr=PIPE, check=True
        )

        assert output == mocked_run().stdout.decode()

    def test_non_zero_exit_code(self, isolated_filesystem):
        with isolated_filesystem:
            with pytest.raises(CalledProcessError):
                _run_command(["git", "status"])


class TestTag:
    @parametrize(
        "text, date, name",
        [
            ("2015-10-15  v10.0.1", "2015-10-15", "v10.0.1"),
            ("2016-02-26   v0.10.13", "2016-02-26", "v0.10.13"),
            ("2016-07-06   save-point", "2016-07-06", "save-point"),
        ],
    )
    def test_parse_name_and_date_string(self, text, date, name):
        tag = Tag(text)

        assert tag.name == name
        assert tag.date == date

    def test_str(self):
        tag = Tag("2015-10-15  v10.0.1")
        assert str(tag) == "v10.0.1"

    def test_repr(self):
        tag = Tag("2015-10-15  v10.0.1")
        assert repr(tag) == 'Tag("2015-10-15  v10.0.1")'


class TestCommitClass:
    @parametrize(
        "short_hash, expected",
        [
            ("eaedb93", "eaedb9320c7aad581daa05f1510b64393c082dbb"),
            ("80a9e0e", "80a9e0edf6208b3c1c9bc28bfb1a5f04b1f59e1d"),
        ],
        ids=[
            "eaedb9320c7aad581daa05f1510b64393c082dbb",
            "80a9e0edf6208b3c1c9bc28bfb1a5f04b1f59e1d",
        ],
    )
    def test_hash(self, commit_text_registry, short_hash, expected):
        text = commit_text_registry[short_hash]
        assert Commit(text).hash == expected

    @parametrize(
        "short_hash, expected",
        [
            ("eaedb93", "Fix lorem ipsum dolor sit amet"),
            ("80a9e0e", "Stop being lazy"),
            ("2f9be68", "Add more lorem ipsum text"),
            ("0a10908", "Add additional information (#26)"),
        ],
        ids=[
            "with body",
            "without line break",
            "without body",
            "with multiple paragraphs",
        ],
    )
    def test_header(self, commit_text_registry, short_hash, expected):
        text = commit_text_registry[short_hash]
        assert Commit(text).header == expected

    @parametrize(
        "short_hash, expected",
        [
            ("4d17c1a", "!fix:thing Fixes #1"),
            ("8c8dcb7", "!refactor:lorem"),
            ("82f0d12", "Sudo make me a sandwich."),
        ],
    )
    def test_footer(self, commit_text_registry, short_hash, expected):
        text = commit_text_registry[short_hash]
        assert Commit(text).footer == expected

    @parametrize(
        "short_hash",
        ["80a9e0e", "2f9be68"],
        ids=["no line break before header", "without body"],
    )
    def test_message_with_invalid_body(self, commit_text_registry, short_hash):
        text = commit_text_registry[short_hash]
        assert Commit(text).body is None

    @parametrize(
        "short_hash, expected",
        [
            (
                "eaedb93",
                (
                    "Consectetur adipiscing elit, sed do\n"
                    "eiusmod tempor incididunt ut labore et\n"
                    "dolore magna aliqua."
                ),
            ),
            ("82f0d12", ("What? Make it yourself.\n" "\n" "Sudo make me a sandwich.")),
            (
                "0a10908",
                (
                    "But I must explain to you how all this mistaken\n"
                    "idea of denouncing pleasure and praising pain was\n"
                    "born.\n"
                    "\n"
                    "Who avoids a pain that produces no resultant pleasure?."
                ),
            ),
        ],
        ids=[
            "Single paragraph",
            "multiple single line paragraphs",
            "multiple paragraphs",
        ],
    )
    def test_messages_with_valid_body(self, commit_text_registry, short_hash, expected):
        text = commit_text_registry[short_hash]
        assert Commit(text).body == expected


class TestGitAdd:
    @parametrize(
        "files, expected",
        [
            (["module.py"], ("module.py",)),
            (["module.py", "folder/file.py"], ("module.py", "folder/file.py")),
        ],
    )
    @patch("braulio.git._run_command", autospec=True)
    def test_run_command_call(self, mocked_run_command, files, expected):
        git = Git()
        git.add(*files)

        command = ("git", "add")
        mocked_run_command.assert_called_with(command + expected)

    def test_when_provided_files_does_not_exits(self):
        git = Git()

        error = "No such file or directory: this_does_not_exist.py"
        with pytest.raises(FileNotFoundError, match=error):
            git.add("this_does_not_exist.py")


class TestGitCommit:
    @patch("braulio.git._run_command", autospec=True)
    def test_call_to_run_command(self, mocked_run_command):
        git = Git()
        git.commit("A message")

        command = ["git", "commit", "-m", '"A message"']
        mocked_run_command.assert_called_with(command)

    @patch("braulio.git._run_command", autospec=True)
    @patch("braulio.git.Git.add", autospec=True)
    def test_pass_files_to_commit(self, mocked_git_add, mocked_run_command):
        git = Git()
        git.commit("A good message", files=["file1.py", "file2.py"])

        # When autospec is used on a class method, the self parameter
        # must be passed to too.
        git.add.assert_called_with(git, "file1.py", "file2.py")

        command = ["git", "commit", "-m", '"A good message"']
        mocked_run_command.assert_called_with(command)


class TestGitLog:
    @patch("braulio.git._run_command", autospec=True)
    def test_log_all_commits(self, mocked_run_command, fake_git_log_output):
        mocked_run_command.return_value = fake_git_log_output

        git = Git()
        commits = git.log()

        mocked_run_command.assert_called_with(["git", "log"])
        assert len(commits) == 12
        assert isinstance(commits[0], Commit)

        first = commits[0]
        assert first.header == "Fix lorem ipsum dolor sit amet"

        last = commits[-1]
        assert last.header == "Add additional information (#26)"

    @parametrize(
        "f, t, revision_range",
        [
            (None, None, []),
            (None, "tag2", []),
            ("tag1", None, ["tag1..HEAD"]),
            ("tag1", "tag2", ["tag1..tag2"]),
        ],
    )
    @patch("braulio.git._run_command", return_value="", autospec=True)
    def test_log_range(self, mocked_run_command, f, t, revision_range):

        git = Git()
        git.log(_from=f, to=t)

        mocked_run_command.assert_called_with(["git", "log"] + revision_range)


git_tag_output = (
    "2015-10-15      v0.0.1\n"
    "2015-11-18      v0.0.2\n"
    "2016-02-26      v0.0.3\n"
    "2016-03-22      v0.0.4\n"
    "2016-07-06      v0.0.5\n"
    "2016-10-03      v0.0.6\n"
)


class TestGitTag:
    @parametrize(
        "command_output, expected", [("", 0), ("  ", 0), ("\n", 0), (git_tag_output, 6)]
    )
    @patch("braulio.git._run_command", autospec=True)
    def test_get_tags(self, mocked_run_command, command_output, expected):
        mocked_run_command.return_value = command_output

        git = Git()
        lst = git.tag()

        mocked_run_command.assert_called_with(
            [
                "git",
                "tag",
                "-l",
                "--sort=creatordate",
                "--format=%(creatordate:short)%09%(refname:strip=2)",
            ]
        )

        assert type(lst) == list
        assert len(lst) == expected

        if expected > 0:
            for _tag in lst:
                assert type(_tag) == Tag

            assert lst[0].name == "v0.0.6"
            assert lst[5].name == "v0.0.1"

    @patch("braulio.git._run_command", autospec=True)
    def test_add_tag(self, mocked_run_command):
        git = Git()
        git.tag("tagname")

        mocked_run_command.assert_called_with(
            ["git", "tag", "-a", "tagname", "-m", '""']
        )


class Test_commit_analyzer:

    # A substitution to braulio.git.Commit for test purpose
    Commit = namedtuple("Commit", ["header", "footer", "message"])

    @parametrize(
        "label_pattern, header",
        [
            ("!{action}:{scope} {subject}", "!fix:cli Make it work"),
            ("[{action}:{scope}] {subject}", "[fix:cli] Make it work"),
            ("{action}({scope}): {subject}", "fix(cli): Make it work"),
            ("{action}({scope}): {subject}", "fix(cli): Make it work"),
        ],
    )
    def test_label_in_header(self, isolated_filesystem, label_pattern, header):

        commit = self.Commit(header=header, footer=None, message="")

        with isolated_filesystem:

            lst = commit_analyzer(
                [commit], label_pattern=label_pattern, label_position="header"
            )

            assert len(lst) == 1
            assert lst[0].subject == "Make it work"
            assert lst[0].scope == "cli"
            assert lst[0].action == "fix"

    @parametrize(
        "label_pattern, footer",
        [
            ("!{action}:{scope}", "!fix:cli"),
            ("[{action}:{scope}]", "[fix:cli]"),
            ("{action}({scope}):", "fix(cli):"),
            ("{action}({scope})", "fix(cli) Fixes #45 and closes #67"),
        ],
    )
    def test_label_in_the_footer(self, isolated_filesystem, label_pattern, footer):

        commit = self.Commit(header="Make it work", footer=footer, message="")

        with isolated_filesystem:

            lst = commit_analyzer(
                [commit], label_pattern=label_pattern, label_position="footer"
            )

            assert len(lst) == 1
            assert lst[0].subject == "Make it work"
            assert lst[0].scope == "cli"
            assert lst[0].action == "fix"

    @parametrize(
        "label_pattern, header",
        [
            ("!{action}:{scope} {subject}", "!fix: Make it work"),
            ("[{action}:{scope}] {subject}", "[fix:] Make it work"),
            ("{action}({scope}): {subject}", "fix(): Make it work"),
        ],
    )
    def test_scopeless_labels(self, isolated_filesystem, label_pattern, header):

        commit = self.Commit(header=header, footer=None, message="")

        with isolated_filesystem:

            lst = commit_analyzer(
                [commit], label_pattern=label_pattern, label_position="header"
            )

            assert len(lst) == 1
            assert lst[0].subject == "Make it work"
            assert lst[0].scope is None
            assert lst[0].action == "fix"

    def test_scopeless_label_pattern(self, isolated_filesystem):

        commit = self.Commit(header="Make it work", footer="!fix", message="")

        with isolated_filesystem:

            lst = commit_analyzer(
                [commit], label_pattern="!{action}", label_position="footer"
            )

            assert len(lst) == 1
            assert lst[0].subject == "Make it work"
            assert lst[0].scope is None
            assert lst[0].action == "fix"



import click
from pathlib import Path
from collections import OrderedDict
from configparser import ConfigParser
from braulio.files import DEFAULT_CHANGELOG


DEFAULT_CONFIG = ConfigParser()
DEFAULT_CONFIG.read_dict(
    {
        "braulio": {
            "commit": "True",
            "message": "Release version {new_version}",
            "tag": "True",
            "confirm": "False",
            "changelog_file": DEFAULT_CHANGELOG,
            "files": "",
            "label_pattern": "!{action}:{scope}",
            "label_position": "footer",
            "tag_pattern": "v{version}",
        },
        "braulio.stages": {"final": "{major}.{minor}.{patch}"},
    }
)


class Config:

    cfg_file_options = {}

    def __init__(self):

        cfg = self._load_config_file()

        self.config_parser = cfg
        self._commit = cfg.getboolean("braulio", "commit")
        self._message = cfg.get("braulio", "message")
        self._tag = cfg.getboolean("braulio", "tag")
        self._confirm = cfg.getboolean("braulio", "confirm")
        self._changelog_file = Path(cfg.get("braulio", "changelog_file").strip())
        self._label_pattern = cfg.get("braulio", "label_pattern").strip()
        self._label_position = cfg.get("braulio", "label_position").strip()
        self._tag_pattern = cfg.get("braulio", "tag_pattern").strip()
        self._current_version = cfg.get("braulio", "current_version", fallback=None)

        self._stages = OrderedDict(cfg["braulio.stages"])

        # files
        files_value = cfg.get("braulio", "files").strip()

        if files_value == "":
            self._files = ()
            return

        if "\n" in files_value:
            files_value = files_value.replace("\n", "")

        file_path_list = files_value.split(",")

        self._files = tuple(fp.strip() for fp in file_path_list)

    def _load_config_file(self):
        path = Path.cwd() / "setup.cfg"
        user_config = ConfigParser()

        if path.exists() and path.is_file():
            user_config.read(path)

            if user_config.has_section("braulio"):
                self.cfg_file_options = OrderedDict(user_config.items("braulio"))

        merged_config = ConfigParser()
        merged_config.read_dict(DEFAULT_CONFIG)

        # Since ConfigParser uses OrderedDict, we need to remove
        # [braulio.stages] default section before merge the user defined
        # section to preserve the order of the options.
        if user_config.has_section("braulio.stages"):
            merged_config.remove_section("braulio.stages")

        merged_config.read_dict(user_config)

        return merged_config

    @property
    def commit(self):
        return self._commit

    @property
    def message(self):
        return self._message

    @property
    def tag(self):
        return self._tag

    @property
    def confirm(self):
        return self._confirm

    @property
    def changelog_file(self):
        return self._changelog_file

    @property
    def files(self):
        return self._files

    @property
    def label_pattern(self):
        return self._label_pattern

    @property
    def label_position(self):
        return self._label_position

    @property
    def tag_pattern(self):
        return self._tag_pattern

    @property
    def current_version(self):
        return self._current_version

    @property
    def stages(self):
        return self._stages


def update_config_file(option, value):
    path = Path.cwd() / "setup.cfg"
    setup_config = ConfigParser()

    if not path.exists():
        path.touch()

    setup_config.read(path)

    if not setup_config.has_section("braulio"):
        setup_config.add_section("braulio")

    setup_config.set("braulio", option, value)

    with path.open("w") as f:
        setup_config.write(f)

    mark = click.style("✓", fg="green", bold=True)
    filename = click.style("setup.cfg", bold=True, fg="blue")
    click.echo(f" {mark} Updated {filename} {option} option with {value}")
